% 公开版代码
% 使用MNIST数据集进行手写字体数字的识别的demo，演示LSTM等方法进行分类的调用方法
% 本程序2021.10定型，后续更新都在完整版代码中
% 本代码地址及代码说明：http://khsci.com/docs/index.php/2021/10/23/rnns_class/
% 注意！公开版代码需使用2022a及以上MATLAB版本，完整版代码可根据代码说明适用相对较低版本MATLAB
close all
clear all
%% 1.加载数据
load MNIST.mat   %载入MNIST数据集
rng(12345);      %设置随机种子，有该行代码可以使每次运行结果保持一致
versionTest;% 代码运行环境监测
%% 2.调用函数进行分类
divideR = [0.6,0.2,0.2];                 %训练集/验证集/测试集比例
hiddenSizes = [32];                      %每个层（如LSTM层）的单元数，如果要设置多个层，则使用数组表示，数值一般在几十到几百之间
                                         % 如第一层隐藏单元数为128，第二层隐藏单元数为64，则设置：hiddenSizes=[128,64]
dropProb = 0;                            % 丢弃层的丢弃概率，如果不丢弃，则设置为0，最大为1
auto = 'on';                             % 是否进行自动纠错，'on'为是，否则为否。开启自动纠错后会智能调整训练、测试集的行列方向。
figflag = 'on';                          % 是否画图，'on'为画图，'off'为不画
rnnName = 'LSTM';                         % 使用的网络名称，可选项包括：'GRU','LSTM','BiLSTM'，注意大小写  ***公开版代码可选'LSTM'，完整版无限制***
options.solverName = 'sgdm';             % 求解器，'sgdm'（默认） | 'rmsprop' | 'adam'                   
options.MaxEpochs = 120;                 % 最大迭代次数，默认30  ***公开版代码最大可设置为200，完整版无限制***
options.InitialLearnRate = 0.01;         % 初始化学习速率（默认0.005）
options.Plots = 'training-progress';     % 是否显示训练过程，'none' 为不显示（默认） | 'training-progress'为显示
options.ValidationFrequency = 10;        % 验证频率，即每间隔多少次迭代进行一次验证
options.LearnRateSchedule = 'piecewise'; % 是否在一定迭代次数后学习速率下降
options.LearnRateDropFactor = 0.8;       % 学习速率下降因子
options.LearnRateDropPeriod = 40;        % 学习速率下降时的迭代数
options.Plots = 'training-progress';     % 显示训练过程
options.GradientThreshold = 1;           % 梯度极限，默认为Inf

% 调用函数执行训练
[accuracy] = FunClassRNNs(TRAIN_images,TRAIN_labels,divideR,rnnName,hiddenSizes,dropProb,auto,options,figflag); % p文件可调用，无法查看源码，无法导出正确率、召回率、精确率和训练好的网络，需要源码请获取完整版代码

% function [accuracy,recall,precision,net] = FunClassRNNs(dataX,dataY,divideR,rnnName,hiddenSizes,dropProb,auto,options,figflag)
% 使用RNN衍生方法（包括GRU、LSTM、BiLSTM）进行模式识别（分类）的快速实现函数
% dataX：输入数据，R*Q的矩阵，R为特征维度，Q为批次数，输入该变量时一定要注意行列方向是否正确
% dataY：标签值，可以为两种方式：
%                              向量型：U*Q的矩阵，U为标签种类数，Q为批次数
%                              索引型：1*Q的矩阵，Q为批次数
% divideR：数据集（训练集、验证集、测试集）划分比例，如：divideR =[0.6,0.2,0.2]，
% 则代表60%数据用于训练集，20%数据用于验证集，20%数据用于测试集
% rnnName：使用的网络名称，可选项包括：'GRU','LSTM','BiLSTM'，注意大小写
% hiddenSizes：每个层（如LSTM层）的单元数，如果要设置多个层，则使用数组表示，数值一般在几十到几百之间
% 如第一层隐藏单元数为100，第二层隐藏单元数为60，则设置：hiddenSizes=[100,20]
% dropProb：丢弃层的丢弃概率，如果不丢弃，则设置为0，最大为1
% auto：是否进行自动纠错，'on'为是，否则为否。开启自动纠错后会智能调整训练、测试集的行列方向。
% options：一些与网络训练等相关的设置，使用结构体方式赋值，比如 options.MaxEpochs = 1000，具体包括：
%             solverName：求解器，'sgdm'（默认） | 'rmsprop' | 'adam'
%             MaxEpochs：最大迭代次数，默认30
%             MiniBatchSize：批尺寸，默认128
%             GradientThreshold：梯度极限，默认为Inf
%             InitialLearnRate：初始化学习速率（默认0.005）
%             Plots：是否显示训练过程，'none' 为不显示（默认） | 'training-progress'为显示
%             ValidationFrequency：验证频率，即每间隔多少次迭代进行一次验证，默认50
%             LearnRateSchedule：即LearnRateSchedule是否在一定迭代次数后学习速率下降， LearnRateSchedule ='piecewise'为使用，'none'为不使用（默认）
%             LearnRateDropPeriod：即LearnRateDropPeriod学习速率下降时的迭代数，默认为10
%             LearnRateDropFactor：即LearnRateDropFactor学习速率下降因子，下降后变为LearnRateDropFactor*InitialLearnRate，LearnRateSchedule为0时可以赋0，默认为0.1
%             （未启用）NorFlag：即Normalization Flag，设置为1时则在程序中进行数据归一化和反归一化操作，否则不进行，建议设置为1
%             SeedFlag：随机种子标志，设置为1时启用随机种子,(默认为1)
% figflag：是否画图，'on'为画图，'off'为不画
% 输出：
% accuracy：测试集分类正确率
% recall： 召回率
% precision： 精确率
% net：训练好的网络

disp(['测试集分类正确率为：',num2str(accuracy)])

%% 关于完整版代码：
% 公开版代码的函数文件为p文件，可以被调用，但无法查看代码。
% 完整版代码中全部为m文件，m文件可以查看源码并自由修改，画图无水印
% 完整版可使用'GRU','LSTM','BiLSTM'三类网络
% 完整版最大迭代次数无限制

% 如果需要封装好的函数的源码，可在下述连接（完整版）获取：
% https://item.taobao.com/item.htm?id=658493893031
% 编程不易，感谢支持~

function versionTest
% 获取MATLAB版本
versionStr = version('-release');
versionYear = str2double(versionStr(1:4));

% 判断MATLAB版本是否为2022a或以上
if (versionYear < 2022)
% 如果版本过低，弹出提示框
warndlg('该代码需使用MATLAB 2022及更新版本运行', '版本警告');
end
end